﻿
using UnityEngine;
using System.Collections;
using System.IO; // for debug save

namespace MsrHmd {

    // A class to intercept the render targets after they have been rendered by
    // each eye camera, perform our own distortion rendering on them, then feed
    // that into the OVR system for presenting.
    public class Sceptor : MonoBehaviour {

        public enum DistortMode {
            PassThru,
            Nullify,
            PreDistort,
            
            Count
        };
        public enum CoefSource {
            
            Coef_5mm,
            Coef_5mm_RGB,
            Line_5mm,
            Line_5mm_RGB,

            Coef_10mm,
            Coef_10mm_RGB,
            Line_10mm,
            Line_10mm_RGB,

            Coef_15mm_03,
            Coef_15mm,
            Coef_15mm_RGB,
            Line_15mm,
            Line_15mm_RGB,

            Coef_OVR,
            Line_OVR,

            Coef_20mm_RGB,
            Line_20mm_RGB,

            Count
        }
        public DistortMode _DistortMode = DistortMode.PreDistort;
        public bool TimeWarp = true;
        public float Amplify { get { return _amplify; } set { _amplify = value; } }
        public float RadAmp { get { return _radAmp; } set { _radAmp = value; } }

        public CoefSource CoefSrc { 
            get { return _CoefSource; } 
            set { 
                _CoefSource = value;
                CreateMeshes();
            } 
        }

        public float LensSeparationMM { 
            get { 
                return lensCenterSeparationMM; 
            } 
            set { 
                lensCenterSeparationMM = value;
                CreateMeshes();
            } 
        }

        public static bool IsEnabledAndActive {
            get {
                bool okay = false;
                Sceptor[] instances = GameObject.FindObjectsOfType<Sceptor>();
                foreach (var inst in instances) {
                    if (inst.enabled) {
                        okay = true;
                    }
                }
                return okay;
            }
        }
        // These are only static to make it convenient for others to get at them
        // (via classname rather than having to find an instance). The end result
        // is the same, because there should be only one Sceptor class instance anyway.
        public static RenderTexture GetIntermediate(OVREye eye) {
            int eyeIdx = (int)eye;
            if (intermediates[eyeIdx] == null) {
                Debug.Log("Intermediate not made for " + eye.ToString() + ", performing setup.");
                SetupIntermediates();
            }
            return intermediates[eyeIdx];
        }
        //public static void SetCamera(OVREye eye, Camera camera) {
        //    _cameras[(int)eye] = camera;
        //}
        private static void SetupIntermediates() {
            for (int eyeIdx = 0; eyeIdx < 2; ++eyeIdx) {
                OVREye eye = (OVREye)eyeIdx;

                OVRDisplay.EyeRenderDesc eyeDesc = OVRManager.display.GetEyeRenderDesc(eye);

                int width = (int)(eyeDesc.resolution.x * OVRManager.instance.nativeTextureScale);
                int height = (int)(eyeDesc.resolution.y * OVRManager.instance.nativeTextureScale);
                int depth = OVRManager.instance.eyeTextureDepth;
                Debug.Log("Sceptor " + eye.ToString() + "W,H,D: " + width + "," + height + "," + depth);
                RenderTextureFormat format = OVRManager.instance.eyeTextureFormat;

                intermediates[eyeIdx] = new RenderTexture(width, height, depth, format);
                intermediates[eyeIdx].antiAliasing = (QualitySettings.antiAliasing == 0) ? 1 : QualitySettings.antiAliasing;
                intermediates[eyeIdx].generateMips = false;
                intermediates[eyeIdx].Create();
                intermediates[eyeIdx].name = "Sceptor " + eye.ToString();
            }
        }
        private static void CheckIntermediates() {
            for (int eyeIdx = 0; eyeIdx < 2; ++eyeIdx) {
                RenderTexture eyeTarg = GetIntermediate((OVREye)eyeIdx);
                if (!eyeTarg.IsCreated()) {
                    eyeTarg.Create();
                }
            }
        }

        private static RenderTexture[] intermediates = new RenderTexture[2];
        //private static Camera[] _cameras = new Camera[2];
        ////////////////////////////////////////////////////////////////////////////

        Material[] _Mats = new Material[(int)DistortMode.Count];

        Mesh[,] _Meshes = new Mesh[(int)DistortMode.Count,2];
        float _amplify = 1.0f;
        float _radAmp = 1.0f;
        private CoefSource _CoefSource = CoefSource.Coef_15mm_03;

        // Use this for initialization
        void Start() {

#if MAF_USE_SETTINGS_FOR_SEP
            // Allow override of the distance between lens centers from the settings file.
            lensCenterSeparationMM = Settings.GetFloat("LensSeparationMM", lensCenterSeparationMM);
#endif // MAF_USE_SETTINGS_FOR_SEP
            if (GameObject.FindObjectOfType<OVRCameraRig>() == null) {
                enabled = false;
            }

            if (enabled) {
                uint toDisable = (uint)Ovr.DistortionCaps.Chromatic; // | (uint)Ovr.DistortionCaps.HqDistortion | (uint)Ovr.DistortionCaps.TimeWarp | (uint)Ovr.DistortionCaps.Vignette;
                if (!TimeWarp) {
                    toDisable |= (uint)Ovr.DistortionCaps.TimeWarp;
                }
                OVRManager.display.distortionCaps = OVRManager.display.distortionCaps & ~toDisable;
                if (!Settings.GetBool("Mirror", true)) {
                    OVRManager.display.mirrorMode = false;
                }

                CreateMaterials();
                CheckIntermediates();
                CreateMeshes();

            }
        }

        private Ovr.Eye Cvt(OVREye eye) {
            switch (eye) {
                case OVREye.Left:
                    return Ovr.Eye.Left;
                case OVREye.Right:
                    return Ovr.Eye.Right;
                default:
                    break;
            }
            Debug.LogError("Bad input to convert from OVREye to Ovr.Eye: " + eye);
            return Ovr.Eye.Count;
        }
        private OVREye Cvt(Ovr.Eye eye) {
            switch (eye) {
                case Ovr.Eye.Left:
                    return OVREye.Left;
                case Ovr.Eye.Right:
                    return OVREye.Right;
                default:
                    break;
            }
            Debug.LogError("Bad input to convert from Ovr.Eye to OVREye: " + eye);
            return OVREye.Center;
        }
        private void CreateMeshes() {
            for (int eyeIdx = 0; eyeIdx < _Meshes.GetLength(1); ++eyeIdx) {
                Ovr.Eye eye = (Ovr.Eye)eyeIdx;
                SetupUVToPixel(eye);
                Ovr.FovPort fovPort = OVRManager.capiHmd.GetDesc().DefaultEyeFov[eyeIdx];
                uint distortionCaps = OVRManager.display.distortionCaps; // | (uint)Ovr.DistortionCaps.Chromatic;
                Ovr.DistortionMesh? ovrMesh = OVRManager.capiHmd.CreateDistortionMesh(eye, fovPort, distortionCaps);
                if (ovrMesh != null) {
                    for (int mode = 0; mode < _Meshes.GetLength(0); ++mode) {
                        _Meshes[mode, eyeIdx] = new Mesh();
                        ConvertMesh(eye, _Meshes[mode, eyeIdx], ovrMesh.Value, (DistortMode)mode);
                    }
                }
                else {
                    Debug.LogError("Unexpected failure on OvrManager.capiHmd.CreateDistortionMesh()");
                }
            }
        }
        private Vector2[] _UVToMMScale = new Vector2[2] { Vector2.one, Vector2.one };
        private Vector2[] _UVToMMOffset = new Vector2[2] { Vector2.zero, Vector2.zero };
        private Vector2[] _RadToUVScale = new Vector2[2] { Vector2.one, Vector2.one };
        private Vector2[] _RadToUVOffset = new Vector2[2] { Vector2.zero, Vector2.zero };
        //private float lensCenterSeparationMM = 63.5f;
        //private float lensCenterSeparationMM = 59.5f;
        private float lensCenterSeparationMM = 67.5f;
        //float sideLensCenterMM = 31.13f; // centered eyecups, distance from nearest side of display to lens center
        //float sideLensCenterMM = 33.13f; // offset eyecups narrow, distance from nearest side of display to lens center
        //float sideLensCenterMM = 29.13f; // offset eyecups wide, distance from nearest side of display to lens center
        private void SetupUVToPixel(Ovr.Eye eye) {
            int eyeIdx = (int)eye;

            float mmDisplayWidth = 125.76f;
            float mmDisplayHeight = 70.74f;
            // Scale to from UV [0..1] to display millimeters [0..dispWidth/2][0..dispHgt]
            _UVToMMScale[eyeIdx] = new Vector2(mmDisplayWidth * 0.5f, mmDisplayHeight);
            // This is based on the assumption that the input UV maps [0..1][0..1] to the entire
            // half of the physical display.
            // (Note that's only left or right half of display).
            // Now we want to convert it to mm distance from eye lens center for display half.
            // mafish!!! Temporary (hopefully) correction for lateral offset of lens in cup. Real distance commented out above.!!!
            float sideLensCenterMM = (mmDisplayWidth - lensCenterSeparationMM) * 0.5f; // offset eyecups narrow, distance from nearest side of display to lens center
            // for left eye, x coordinate of center is sideLensCenter, but for right eye
            // zero corresponds to physical display center, so x coord is 
            // half physical display width - sideLensCenter
            float leftLensCenterMM = eye == Ovr.Eye.Left
                ? sideLensCenterMM
                : (mmDisplayWidth * 0.5f - sideLensCenterMM);
            float topLensCenterMM = 35.37f;
            _UVToMMOffset[eyeIdx] = new Vector2(-leftLensCenterMM, -topLensCenterMM);
            // Now uv=[0..1] * _UVToMMScale + _UVToMMOffset
            //      = mm distance from lens center

            // Apparently the internal system is not using the entire rendered image.
            // That is, rather than:
            //      fov = max(fovPort.minTan, fovPort.maxTan) * 2;
            //      uv[0..1] == tan(rads) / tan(fov/2) * 0.5f + 0.5f
            // we have:
            //      uv[0..1] = ( tan(rads) - fovPort.minTan ) / ( fovPort.maxTan - fovPort.minTan );
            Ovr.FovPort fovPort = OVRManager.capiHmd.GetDesc().DefaultEyeFov[eyeIdx];
            Ovr.EyeRenderDesc eyeDesc = OVRManager.capiHmd.GetRenderDesc(eye, fovPort);
            // Here fovPort.RightTan and fovPort.DownTan are maxTan, and
            // -fovPort.LeftTan and -fovPort.UpTan are minTan. NOTE the negation, all fovPort.tan values are positive!!!!
            Vector2 maxTan = new Vector2(fovPort.RightTan, fovPort.DownTan);
            Vector2 minTan = new Vector2(-fovPort.LeftTan, -fovPort.UpTan);
            _RadToUVScale[eyeIdx] = new Vector2(1.0f / (maxTan.x - minTan.x), 1.0f / (maxTan.y - minTan.y));
            _RadToUVOffset[eyeIdx] = Vector2.Scale(-minTan,  _RadToUVScale[eyeIdx]);
        }
        private Vector2 UVToMMFromCenter(Ovr.Eye eye, Vector2 uv) {
            int eyeIdx = (int)eye;
            return Vector2.Scale(uv, _UVToMMScale[eyeIdx]) + _UVToMMOffset[eyeIdx];
        }
        private Vector2 UVFromRadians(Ovr.Eye eye, Vector2 dir, float rads) {
            int eyeIdx = (int)eye;
#if MAF_UV_AS_RAW_RADS
            return new Vector2(rads * dir.x, rads * dir.y);
#else // MAF_UV_AS_RAW_RADS
            float tanRads = Mathf.Tan(rads);
            Vector2 tanRadsVec = new Vector2(dir.x * tanRads, dir.y * tanRads);
            return Vector2.Scale(tanRadsVec, _RadToUVScale[eyeIdx]) + _RadToUVOffset[eyeIdx];
#endif // MAF_UV_AS_RAW_RADS
        }
        private enum ColorBand {
            Red,
            Green,
            Blue
        };
        static float[] coef_5mm = new float[5] { 0.0f, 24.819f, -129.75f, 7002.2f, -105782.0f };
        static float[] coef_10mm = new float[5] { 0.0f, 24.479f, -119.92f, 8435.9f, -152205.0f };
        static float[] coef_15mm = new float[5] { 0.0f, 24.121f, -93.366f, 8475.3f, -178944.0f };
        static float[] coef_15mm_03 = new float[5] { 0.0f, 24.84463f, -96.16698f, 8729.559f, -184312.32f };
        //float[] coef_OVR = new float[5] { 0.0f, 28.847f, -235.12f, 6547.6f, -68750.0f };
        static float[] coef_OVR = new float[5] { 0.0f, 29.257f, -235.04f, 6813.6f, -74860.0f };

        static float[] line_5mm = new float[5] { 0.0f, 24.14f, 0.0f, 0.0f, 0.0f };
        static float[] line_10mm = new float[5] { 0.0f, 24.267f, 0.0f, 0.0f, 0.0f };
        static float[] line_15mm = new float[5] { 0.0f, 24.185f, 0.0f, 0.0f, 0.0f };
        static float[] line_OVR = new float[5] { 0.0f, 25.732f, 0.0f, 0.0f, 0.0f };

        static float[] coef_5mm_R = new float[5] { 0.0f, 25.003f, -155.32f, 7300.6f, -84521.0f };
        static float[] coef_5mm_G = new float[5] { 0.0f, 25.04f, -157.91f, 7458.9f, -87605.0f };
        static float[] coef_5mm_B = new float[5] { 0.0f, 25.24f, -184.28f, 8780.5f, -108496.0f };
        static float[][] coef_5mm_RGB = new float[3][] { coef_5mm_R, coef_5mm_G, coef_5mm_B };

        static float[] coef_10mm_R = new float[5] { 0.0f, 24.54f, -138.49f, 9270.5f, -154782.0f };
        static float[] coef_10mm_G = new float[5] { 0.0f, 24.616f, -149.13f, 9919.1f, -167431.0f };
        static float[] coef_10mm_B = new float[5] { 0.0f, 24.707f, -158.98f, 10555.0f, -182064.0f };
        static float[][] coef_10mm_RGB = new float[3][] { coef_10mm_R, coef_10mm_G, coef_10mm_B };

        static float[] coef_15mm_R = new float[5] { 0.0f, 24.102f, -97.84f, 8667.3f, -174225.0f };
        static float[] coef_15mm_G = new float[5] { 0.0f, 24.134f, -99.742f, 8837.0f, -179989.0f };
        static float[] coef_15mm_B = new float[5] { 0.0f, 24.222f, -111.58f, 9747.3f, -205138.0f };
        static float[][] coef_15mm_RGB = new float[3][] { coef_15mm_R, coef_15mm_G, coef_15mm_B };

        static float[] coef_20mm_R = new float[5] { 0.0f, 23.766f, -62.496f, 7459.3f, -159559.0f };
        static float[] coef_20mm_G = new float[5] { 0.0f, 23.8f, -64.852f, 7693.5f, -168569.0f };
        static float[] coef_20mm_B = new float[5] { 0.0f, 23.868f, -73.305f, 8475.6f, -196826.0f };
        static float[][] coef_20mm_RGB = new float[3][] { coef_20mm_R, coef_20mm_G, coef_20mm_B };

        static float[] line_5mm_R = new float[] { 0.0f, 24.714f };
        static float[] line_5mm_G = new float[] { 0.0f, 24.722f };
        static float[] line_5mm_B = new float[] { 0.0f, 24.721f };
        static float[][] line_5mm_RGB = new float[][] { line_5mm_R, line_5mm_G, line_5mm_B };

        static float[] line_10mm_R = new float[] { 0.0f, 24.399f };
        static float[] line_10mm_G = new float[] { 0.0f, 24.403f };
        static float[] line_10mm_B = new float[] { 0.0f, 24.389f };
        static float[][] line_10mm_RGB = new float[][] { line_10mm_R, line_10mm_G, line_10mm_B };

        static float[] line_15mm_R = new float[] { 0.0f, 24.218f };
        static float[] line_15mm_G = new float[] { 0.0f, 24.226f };
        static float[] line_15mm_B = new float[] { 0.0f, 24.213f };
        static float[][] line_15mm_RGB = new float[][] { line_15mm_R, line_15mm_G, line_15mm_B };
        
        static float[] line_20mm_R = new float[] { 0.0f, 24.101f };
        static float[] line_20mm_G = new float[] { 0.0f, 24.112f };
        static float[] line_20mm_B = new float[] { 0.0f, 24.112f };
        static float[][] line_20mm_RGB = new float[][] { line_20mm_R, line_20mm_G, line_20mm_B };

        private float[] GetCoefs(ColorBand band) {

            switch (_CoefSource) {
                case CoefSource.Coef_5mm:
                    return coef_5mm;
                case CoefSource.Coef_10mm:
                    return coef_10mm;
                case CoefSource.Coef_15mm:
                    return coef_15mm;
                case CoefSource.Coef_15mm_03:
                    return coef_15mm_03;
                case CoefSource.Coef_OVR:
                    return coef_OVR;

                case CoefSource.Line_5mm:
                    return line_5mm;
                case CoefSource.Line_10mm:
                    return line_10mm;
                case CoefSource.Line_15mm:
                    return line_15mm;
                case CoefSource.Line_OVR:
                    return line_OVR;

                case CoefSource.Coef_5mm_RGB:
                    return coef_5mm_RGB[(int)band];
                case CoefSource.Coef_10mm_RGB:
                    return coef_10mm_RGB[(int)band];
                case CoefSource.Coef_15mm_RGB:
                    return coef_15mm_RGB[(int)band];
                case CoefSource.Coef_20mm_RGB:
                    return coef_20mm_RGB[(int)band];

                case CoefSource.Line_5mm_RGB:
                    return line_5mm_RGB[(int)band];
                case CoefSource.Line_10mm_RGB:
                    return line_10mm_RGB[(int)band];
                case CoefSource.Line_15mm_RGB:
                    return line_15mm_RGB[(int)band];
                case CoefSource.Line_20mm_RGB:
                    return line_20mm_RGB[(int)band];

                default:
                    Debug.LogError("Bad coefficent source " + _CoefSource.ToString());
                    break;
            }
            return coef_5mm;
        }
        private float ExpandPolynomial(float rMM, ColorBand band) {
            float rM = rMM * 1.0e-3f;
            float[] coefs = GetCoefs(band);
            //float cRads = (((coefs[4] * rM + coefs[3]) * rM + coefs[2]) * rM + coefs[1]) * rM + coefs[0];
            //return cRads;
            float accum = 0.0f;
            for (int iCoef = coefs.Length - 1; iCoef >= 0; --iCoef) {
                accum *= rM;
                accum += coefs[iCoef];
            }
            return accum;
        }
        private Vector2 GetCorrectedUV(Ovr.Eye eye, Vector2 uv, ColorBand band) {
            Vector2 dir = UVToMMFromCenter(eye, uv);
            float mmRadius = dir.magnitude;
            if (mmRadius > 0) {
                dir /= mmRadius;
            }
            float radiusRads = ExpandPolynomial(mmRadius, band);
            return UVFromRadians(eye, dir, radiusRads); 
        }
        private void GetCorrectedUVs(Ovr.Eye eye, Vector2 orig, out Vector2 uvR, out Vector2 uvG, out Vector2 uvB) {
            uvR = GetCorrectedUV(eye, orig, ColorBand.Red);
            uvG = GetCorrectedUV(eye, orig, ColorBand.Green);
            uvB = GetCorrectedUV(eye, orig, ColorBand.Blue);

            //// mafish
            //uvR = orig;
            //uvG = orig;
            //uvB = orig;
        }
        private Vector3 ExpandPosition(Ovr.Eye eye, Ovr.DistortionVertex srcVtx) {
            Vector3 pos;
            if (eye == Ovr.Eye.Left) {
                pos = new Vector3(srcVtx.ScreenPosNDC.x * 2.0f + 1.0f, srcVtx.ScreenPosNDC.y, srcVtx.TimeWarpFactor);
            }
            else {
                pos = new Vector3(srcVtx.ScreenPosNDC.x * 2.0f - 1.0f, srcVtx.ScreenPosNDC.y, srcVtx.TimeWarpFactor);
            }
            return pos;
        }
        private Vector2 ExpandPosition2(Ovr.Eye eye, Ovr.DistortionVertex srcVtx) {
            Vector2 pos;
            if (eye == Ovr.Eye.Left) {
                pos = new Vector2(srcVtx.ScreenPosNDC.x * 2.0f + 1.0f, srcVtx.ScreenPosNDC.y);
            }
            else {
                pos = new Vector2(srcVtx.ScreenPosNDC.x * 2.0f - 1.0f, srcVtx.ScreenPosNDC.y);
            }
            return pos;
        }
        private Vector4 GetScaleAndOffset(OVREye eye, RenderTexture eyeTarg) {
            int eyeIdx = (int)eye;
            Ovr.FovPort fovPort = OVRManager.capiHmd.GetDesc().DefaultEyeFov[eyeIdx];
            Ovr.Sizei size = new Ovr.Sizei(eyeTarg.width, eyeTarg.height);
            Ovr.Recti viewport = new Ovr.Recti() {
                Pos = new Ovr.Vector2i(0, 0),
                Size = size
            };
            Ovr.Vector2f[] sNo = OVRManager.capiHmd.GetRenderScaleAndOffset(fovPort, size, viewport);

            return new Vector4(sNo[0].x, sNo[0].y, sNo[1].x, sNo[1].y);
        }
        private Vector2 ApplyScaleAndOffset(Vector4 scaleOff, Vector2 uv) {
            return new Vector2(uv.x * scaleOff.x + scaleOff.z, uv.y * scaleOff.y + scaleOff.w);
        }
        private void NullifyOVR(Vector2 ovrPos, Vector2 ovrUV, out Vector2 newPos, out Vector2 newUV) {
            newPos = ovrUV * 2.0f - Vector2.one;
            newUV = ovrPos * 0.5f + new Vector2(0.5f, 0.5f);
        }
        private void ConvertMesh(Ovr.Eye eye, Mesh dstMesh, Ovr.DistortionMesh srcMesh, DistortMode mode) {
            uint numVerts = srcMesh.VertexCount;
            // Would like to use mesh.colors as a holder for 4 floats, but we don't trust them
            // not to convert them to 4 bytes under the hood. %(
            Vector3[] positions = new Vector3[numVerts];
            Vector3[] normals = new Vector3[numVerts];
            Vector2[] uv = new Vector2[numVerts];
            Vector2[] uv2 = new Vector2[numVerts];
            Color[] colors = new Color[numVerts]; // Debug only, remove at earliest convenience.

            int eyeIdx = (int)eye;
            Vector4 scaleOff = GetScaleAndOffset(Cvt(eye), GetIntermediate(Cvt(eye)));
            //Vector4 scaleOff = new Vector4(_RadToUVScale[eyeIdx].x, _RadToUVScale[eyeIdx].y, _RadToUVOffset[eyeIdx].x, _RadToUVOffset[eyeIdx].y);

            for (int iVert = 0; iVert < numVerts; ++iVert) {
                Ovr.DistortionVertex srcVtx = srcMesh.pVertexData[iVert];
                if (mode == DistortMode.PreDistort) {
                    Vector2 ovrPos = ExpandPosition2(eye, srcVtx);
                    Vector2 ovrUV = ApplyScaleAndOffset(scaleOff, srcVtx.TanEyeAnglesR.ToVector2());

                    Vector2 newPos;
                    Vector2 newUV;
                    NullifyOVR(ovrPos, ovrUV, out newPos, out newUV);
                    Vector2 uvR;
                    Vector2 uvG;
                    Vector2 uvB;
                    GetCorrectedUVs(eye, newUV, out uvR, out uvG, out uvB);

                    positions[iVert] = new Vector3(newPos.x, newPos.y, srcVtx.TimeWarpFactor);
                    normals[iVert] = new Vector3(uvR.x, uvR.y, srcVtx.VignetteFactor);
                    uv[iVert] = uvG;
                    uv2[iVert] = uvB;
                //normals[iVert].x = (uvR * 2.0f - new Vector2(1.0f, 1.0f)).magnitude; // mafish
                }
                else if(mode == DistortMode.Nullify) {
                    Vector2 ovrPos = ExpandPosition2(eye, srcVtx);
                    Vector2 ovrUV = ApplyScaleAndOffset(scaleOff, srcVtx.TanEyeAnglesR.ToVector2());

                    Vector2 newPos;
                    Vector2 newUV;
                    NullifyOVR(ovrPos, ovrUV, out newPos, out newUV);
                    Vector2 uvR = newUV;
                    Vector2 uvG = newUV;
                    Vector2 uvB = newUV;

                    positions[iVert] = new Vector3(newPos.x, newPos.y, srcVtx.TimeWarpFactor);
                    normals[iVert] = new Vector3(uvR.x, uvR.y, srcVtx.VignetteFactor);
                    uv[iVert] = uvG;
                    uv2[iVert] = uvB;

                }
                else if(mode == DistortMode.PassThru) {
                    Vector2 ovrPos = ExpandPosition2(eye, srcVtx);
                    Vector2 ovrUV = ovrPos * 0.5f + new Vector2(0.5f, 0.5f);
                    ovrUV.y = 1.0f - ovrUV.y;

                    positions[iVert] = new Vector3(ovrPos.x, ovrPos.y, srcVtx.TimeWarpFactor);
                    normals[iVert] = new Vector3(ovrUV.x, ovrUV.y, srcVtx.VignetteFactor);
                    uv[iVert] = ovrUV;
                    uv2[iVert] = ovrUV;
                }

                // Debug visualization
                float grey = (float)iVert / (float)(numVerts-1);
                colors[iVert] = new Color(grey, grey, grey, 1.0f);
            }
            dstMesh.vertices = positions;
            dstMesh.normals = normals;
            dstMesh.colors = colors;
            dstMesh.uv = uv;
            dstMesh.uv2 = uv2;

            uint numIndices = srcMesh.IndexCount;
            int[] indices = new int[numIndices];
            for (uint idx = 0; idx < numIndices; ++idx) {
                indices[idx] = srcMesh.pIndexData[idx];
            }
            dstMesh.triangles = indices;
        }

        int updtFrameIdx = 0;
        // Update is called once per frame
        void Update() {
            CheckIntermediates();
            if (Input.GetKeyDown(KeyCode.J)) {
                DebugSave(0);
                DebugSave(1);
            }
            updtFrameIdx++;
        }

        private void SetupMat(int eyeIdx, RenderTexture src) {
            Material mat = _Mats[(int)_DistortMode];
            switch (_DistortMode) {
                case DistortMode.PreDistort: {
#if MAF_UV_AS_RAW_RADS
                        Vector4 scaleAndOffset = new Vector4(
                            _RadToUVScale[eyeIdx].x * Amplify,
                            _RadToUVScale[eyeIdx].y * Amplify,
                            _RadToUVOffset[eyeIdx].x,
                            _RadToUVOffset[eyeIdx].y);
                        mat.SetVector("_ScaleOffset", scaleAndOffset);
                        mat.SetColor("_Color", new Color(1.0f, 1.0f, 1.0f, _radAmp));
#endif // MAF_UV_AS_RAW_RADS
                    }
                    break;
                case DistortMode.Nullify:
                case DistortMode.PassThru: { 
                        Vector4 scaleAndOffset = new Vector4(1.0f, 1.0f, 0.0f, 0.0f);
                        mat.SetVector("_ScaleOffset", scaleAndOffset);
                    }
                    break;
                default:
                    Debug.LogError("Bad distortmode " + _DistortMode.ToString());
                    break;
            }
            mat.SetTexture("_MainTex", src);
            mat.SetPass(0);
        }
        static int guiFrameIdx = 0; // for debug save
        //void OnPostRender() {
        void OnGUI() {
            // Important to keep from skipping render events
            if (Event.current.type != EventType.Repaint)
                return;

            for (int eyeIdx = 0; eyeIdx < 2; ++eyeIdx) {

                OVREye eye = (OVREye)eyeIdx;
                RenderTexture src = GetIntermediate(eye);
                RenderTexture dst = OVRManager.display.GetEyeTexture(eye);

                RenderTexture restore = RenderTexture.active;
                RenderTexture.active = dst;
                //GL.Clear(true, true, new Color(1.0f, 1.0f, 0.5f, 1.0f));
                GL.Clear(true, true, new Color(0.0f, 0.0f, 0.0f, 1.0f));
                SetupMat(eyeIdx, src);

                DrawMesh(eyeIdx);
                RenderTexture.active = restore;

                DebugSave(eyeIdx);

            }
            ++guiFrameIdx;
        }

        private void DrawMesh(int eyeIdx) {
            GL.PushMatrix();
            GL.LoadIdentity();
            //GL.wireframe = true; // mafish
            Graphics.DrawMeshNow(_Meshes[(int)_DistortMode, eyeIdx], Matrix4x4.identity);
            //GL.wireframe = false; // mafish
            GL.PopMatrix();
        }

        private void CreateMaterials() {
            //Shader shader = Shader.Find("MsrHmd/Blit");
            for (int iMat = 0; iMat < _Mats.Length; ++iMat) {
#if MAF_UV_AS_RAW_RADS
                Shader shader = iMat == (int)DistortMode.PreDistort
                    ? Shader.Find("MsrHmd/DistMeshRads")
                    : Shader.Find("MsrHmd/DistMesh");
#else //MAF_UV_AS_RAW_RADS
                Shader shader = Shader.Find("MsrHmd/DistMesh");
#endif // MAF_UV_AS_RAW_RADS
                Material mat = new Material(shader);
                mat.hideFlags = HideFlags.HideAndDontSave;
                _Mats[iMat] = mat;
            }
        }

        private void DebugSave(int eyeIdx) {
            if (guiFrameIdx == 0) {
                OVREye eye = (OVREye)eyeIdx;
                RenderTexture restoreTarget = RenderTexture.active;
                //RenderTexture.active = GetIntermediate(eye);
                RenderTexture.active = OVRManager.display.GetEyeTexture(eye);

                int width = RenderTexture.active.width;
                int height = RenderTexture.active.height;

                // Create a destination copy. ReadPixels reads from currently set RenderTarget into destination.
                Texture2D dstTex = new Texture2D(width, height);
                dstTex.ReadPixels(new Rect(0, 0, width, height), 0, 0, false);
                // Apply makes it actually happen.
                dstTex.Apply();

                // If you haven't written all pixels, or some of your shaders are emitting transparent pixels,
                // then the alpha values in the frame might not be set, and those pixels will get trashed
                // on the conversion to PNG.
                bool SetAlpha = true;
                if (SetAlpha) {
                    // Now set all alpha values to opaque. You might not want to do this, or you might want to do
                    // it more efficiently.
                    Color32[] pixels = dstTex.GetPixels32();
                    for (int i = 0; i < pixels.Length; ++i) {
                        pixels[i].a = 255;
                    }
                    // Technically, you should call dstTex.Apply() after the SetPixels, but we don't really need to
                    // ship the new pixels back up to the GPU, we just want them on the CPU to write. But if the
                    // alpha isn't getting set to one, the missing Apply() might be the problem.
                    dstTex.SetPixels32(pixels);
                }

                // Now encode as PNG and write it out.
                byte[] dstBytes = dstTex.EncodeToJPG();
                File.WriteAllBytes(Application.dataPath + "/../dstTex_"
                    + eyeIdx.ToString() + "_"
                    + guiFrameIdx.ToString("D4") + ".jpg", dstBytes);

                RenderTexture.active = restoreTarget;
            }
        }
    }

};